/* RISCOS FILETYPES
 * These are all from the big list at
 * http://wombat.doc.ic.ac.uk/acorn/doc/filetypes
 *
 * To cut down on size, 
 * - only officially registered filetypes are used.
 * - of them, only the "Short name" is used.
 *
 * This code is in the Public Domain.
 */

struct RISCOSfiletype {
  UWORD  id;
  STRPTR name;
};

static const struct RISCOSfiletype RISCOSfiletypes[] = {
  /* most regular types come first */
  {0xF95, "Code"},
  {0xFD7, "TaskObey"},
  {0xFEB, "Obey"},
  {0xFF8, "Absolute"},
  {0xFF9, "Sprite"},
  {0xFFA, "Module"},
  {0xFFB, "BASIC"},
  {0xFFC, "Utility"},
  {0xFFD, "Data"},
  {0xFFE, "Command"},
  {0xFFF, "Text"},

  {0x100, "1stsheet"},
  {0x101, "Calcsprd"},
  {0x102, "Perl"},
  {0x103, "UniMode"},
  {0x107, "OreDBase"},
  {0x108, "DataVox"},
  {0x118, "Angband"},
  {0x11A, "Z-Code"},
  {0x11D, "TAP"},
  {0x122, "Genetic"},
  {0x127, "MOPAC"},
  {0x128, "MOL"},
  {0x129, "Alchemy"},
  {0x12A, "PDB"},
  {0x12B, "RasMolSc"},
  {0x138, "MapprMap"},
  {0x152, "PolyData"},
  {0x153, "Cabri"},
  {0x156, "PsiWord"},
  {0x157, "PsiSheet"},
  {0x158, "Psion"},
  {0x159, "PsiRecrd"},
  {0x15A, "PsiOPL"},
  {0x15B, "PsiData"},
  {0x15C, "PsiAgnda"},
  {0x16D, "Organizer"},
  {0x187, "awk"},
  {0x18C, "Lua"},
  {0x18D, "TaskLua"},
  {0x3E2, "VistTune"},
  {0x9F8, "HD_BackJ"},
  {0xAAA, "RiscOS3"},
  {0xAC9, "VF"},
  {0xACA, "LaTeX"},
  {0xAD8, "PsiMBM"},
  {0xADF, "PDF"},
  {0xAED, "SheetMac"},
  {0xAEE, "SheetWS"},
  {0xAEF, "Sheet"},
  {0xAF0, "ARCWrite"},
  {0xAF1, "Music"},
  {0xAF2, "PostTray"},
  {0xAF3, "PostBox"},
  {0xAF4, "PostFile"},
  {0xAF5, "PostData"},
  {0xAF6, "SimTrace"},
  {0xAFF, "DrawFile"},
  {0xB02, "Meta"},
  {0xB78, "MIDIWrks"},
  {0xB11, "TTXscrpt"},
  {0xB12, "TTXpage"},
  {0xB22, "DFSImage"},
  {0xB30, "FuzzySheet"},
  {0xB31, "FuzzySheet"},
  {0xB3A, "SpellFrq"},
  {0xB3B, "SpellBnk"},
  {0xB4F, "Alphabet"},
  {0xB50, "RGSCOpts"},
  {0xB6A, "Illusion"},
  {0xB6F, "WTModule"},
  {0xB72, "Compo"},
  {0xB7B, "Life"},
  {0xB7C, "LifeForm"},
  {0xB8C, "Split"},
  {0xB95, "WimpBasic"},
  {0xBA0, "Z80"},
  {0xBA1, "SNA"},
  {0xBB0, "BibleNts"},
  {0xBC5, "ImpDoc"},
  {0xBD6, "Audiowrk"},
  {0xC24, "Architech"},
  {0xC35, "CorlDraw"},
  {0xC46, "UnixTar"},
  {0xC7B, "PhoneNum"},
  {0xCA0, "Metafont"},
  {0xCA1, "TFM"},
  {0xCA2, "GF"},
  {0xCA3, "PK"},
  {0xCB5, "STSample"},
  {0xCB6, "STModule"},
  {0xCD5, "Tabs"},
  {0xCE5, "TeX"},
  {0xD93, "DDF"},
  {0xF89, "GZip"},
  {0xF9D, "DiscCD"},
  {0xF9E, "DiscDP"},
  {0xF9F, "DiscD"},
  {0xFB0, "Allocate"},
  {0xFB2, "MovieFS"},
  {0xFB4, "DiscR"},
  {0xFB5, "NoDisc"},
  {0xFBC, "PBAbacus"},
  {0xFBD, "PBWrite"},
  {0xFBE, "PBCards"},
  {0xFC2, "AIFF"},
  {0xFC9, "WiniUtil"},
  {0xFD0, "PCConfig"},
  {0xFD3, "DDEDebug"},
  {0xFD4, "MIDI"},
  {0xFD5, "RiscOS3"},
  {0xFDD, "MasterUt"},
  {0xFDE, "VTSetup"},
  {0xFDF, "VTScript"},
  {0xFE0, "Accessry"},
  {0xFE1, "Makefile"},
  {0xFEA, "Desktop"},
  {0xFEC, "Template"},
  {0xFED, "Palette"},
  {0xFEE, "NotePad"},
  {0xFEF, "Diary"},
  {0xFF0, "TIFF"},
  {0xFF1, "RawData"},
  {0xFF2, "Config"},
  {0xFF3, "LaserJet"},
  {0xFF4, "Printout"},
  {0xFF5, "PoScript"},
  {0xFF6, "Font"},
  {0xFF7, "BBCFont"},
  {0x000, NULL}
};

static STRPTR GetRISCOSfiletype(UWORD id) {
  if ((id >= 0x100) && (id <= 0xFFF)) {
    const struct RISCOSfiletype *ft;
    for (ft = &RISCOSfiletypes[0]; ft->name != NULL; ft++) {
      if (ft->id == id) return ft->name;
    }
  }
  return NULL;
}
